# Generated from ../flash_patcher/antlr_source/PatchfileParser.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,20,99,2,0,7,0,2,1,7,1,2,2,7,2,2,3,7,3,2,4,7,4,2,5,7,5,2,6,7,
        6,2,7,7,7,2,8,7,8,2,9,7,9,1,0,1,0,1,0,1,0,1,1,4,1,26,8,1,11,1,12,
        1,27,1,1,1,1,1,1,1,1,1,2,4,2,35,8,2,11,2,12,2,36,1,3,1,3,1,3,1,3,
        1,3,1,3,1,4,1,4,1,4,1,4,1,5,4,5,50,8,5,11,5,12,5,51,1,5,1,5,1,5,
        1,5,1,5,1,5,1,5,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,6,1,7,4,7,71,8,
        7,11,7,12,7,72,1,8,1,8,1,8,1,8,5,8,79,8,8,10,8,12,8,82,9,8,1,9,3,
        9,85,8,9,1,9,1,9,1,9,3,9,90,8,9,1,9,3,9,93,8,9,1,9,1,9,3,9,97,8,
        9,1,9,0,0,10,0,2,4,6,8,10,12,14,16,18,0,0,101,0,20,1,0,0,0,2,25,
        1,0,0,0,4,34,1,0,0,0,6,38,1,0,0,0,8,44,1,0,0,0,10,49,1,0,0,0,12,
        60,1,0,0,0,14,70,1,0,0,0,16,80,1,0,0,0,18,96,1,0,0,0,20,21,5,1,0,
        0,21,22,5,5,0,0,22,23,3,18,9,0,23,1,1,0,0,0,24,26,3,0,0,0,25,24,
        1,0,0,0,26,27,1,0,0,0,27,25,1,0,0,0,27,28,1,0,0,0,28,29,1,0,0,0,
        29,30,5,6,0,0,30,31,3,4,2,0,31,32,5,17,0,0,32,3,1,0,0,0,33,35,5,
        18,0,0,34,33,1,0,0,0,35,36,1,0,0,0,36,34,1,0,0,0,36,37,1,0,0,0,37,
        5,1,0,0,0,38,39,5,2,0,0,39,40,5,5,0,0,40,41,3,18,9,0,41,42,5,13,
        0,0,42,43,3,18,9,0,43,7,1,0,0,0,44,45,5,3,0,0,45,46,5,5,0,0,46,47,
        3,18,9,0,47,9,1,0,0,0,48,50,3,8,4,0,49,48,1,0,0,0,50,51,1,0,0,0,
        51,49,1,0,0,0,51,52,1,0,0,0,52,53,1,0,0,0,53,54,5,7,0,0,54,55,3,
        14,7,0,55,56,5,19,0,0,56,57,5,6,0,0,57,58,3,4,2,0,58,59,5,17,0,0,
        59,11,1,0,0,0,60,61,5,4,0,0,61,62,5,5,0,0,62,63,5,7,0,0,63,64,3,
        14,7,0,64,65,5,19,0,0,65,66,5,6,0,0,66,67,3,4,2,0,67,68,5,17,0,0,
        68,13,1,0,0,0,69,71,5,20,0,0,70,69,1,0,0,0,71,72,1,0,0,0,72,70,1,
        0,0,0,72,73,1,0,0,0,73,15,1,0,0,0,74,79,3,2,1,0,75,79,3,6,3,0,76,
        79,3,10,5,0,77,79,3,12,6,0,78,74,1,0,0,0,78,75,1,0,0,0,78,76,1,0,
        0,0,78,77,1,0,0,0,79,82,1,0,0,0,80,78,1,0,0,0,80,81,1,0,0,0,81,17,
        1,0,0,0,82,80,1,0,0,0,83,85,5,10,0,0,84,83,1,0,0,0,84,85,1,0,0,0,
        85,86,1,0,0,0,86,87,5,8,0,0,87,89,5,14,0,0,88,90,5,12,0,0,89,88,
        1,0,0,0,89,90,1,0,0,0,90,92,1,0,0,0,91,93,5,11,0,0,92,91,1,0,0,0,
        92,93,1,0,0,0,93,97,1,0,0,0,94,97,5,12,0,0,95,97,5,9,0,0,96,84,1,
        0,0,0,96,94,1,0,0,0,96,95,1,0,0,0,97,19,1,0,0,0,10,27,36,51,72,78,
        80,84,89,92,96
    ]

class PatchfileParser ( Parser ):

    grammarFileName = "PatchfileParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'('", "')'", "<INVALID>", 
                     "'-'" ]

    symbolicNames = [ "<INVALID>", "ADD", "REMOVE", "REPLACE", "REPLACE_ALL", 
                      "FILENAME", "BEGIN_PATCH", "BEGIN_CONTENT", "FUNCTION", 
                      "END", "OPEN_BLOCK", "CLOSE_BLOCK", "INTEGER", "DASH", 
                      "FUNCTION_NAME", "WHITESPACE", "COMMENT", "END_PATCH", 
                      "AS_TEXT", "END_CONTENT", "CONTENT_TEXT" ]

    RULE_addBlockHeader = 0
    RULE_addBlock = 1
    RULE_addBlockText = 2
    RULE_removeBlock = 3
    RULE_replaceNthBlockHeader = 4
    RULE_replaceNthBlock = 5
    RULE_replaceAllBlock = 6
    RULE_replaceBlockText = 7
    RULE_root = 8
    RULE_locationToken = 9

    ruleNames =  [ "addBlockHeader", "addBlock", "addBlockText", "removeBlock", 
                   "replaceNthBlockHeader", "replaceNthBlock", "replaceAllBlock", 
                   "replaceBlockText", "root", "locationToken" ]

    EOF = Token.EOF
    ADD=1
    REMOVE=2
    REPLACE=3
    REPLACE_ALL=4
    FILENAME=5
    BEGIN_PATCH=6
    BEGIN_CONTENT=7
    FUNCTION=8
    END=9
    OPEN_BLOCK=10
    CLOSE_BLOCK=11
    INTEGER=12
    DASH=13
    FUNCTION_NAME=14
    WHITESPACE=15
    COMMENT=16
    END_PATCH=17
    AS_TEXT=18
    END_CONTENT=19
    CONTENT_TEXT=20

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class AddBlockHeaderContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ADD(self):
            return self.getToken(PatchfileParser.ADD, 0)

        def FILENAME(self):
            return self.getToken(PatchfileParser.FILENAME, 0)

        def locationToken(self):
            return self.getTypedRuleContext(PatchfileParser.LocationTokenContext,0)


        def getRuleIndex(self):
            return PatchfileParser.RULE_addBlockHeader

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAddBlockHeader" ):
                listener.enterAddBlockHeader(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAddBlockHeader" ):
                listener.exitAddBlockHeader(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAddBlockHeader" ):
                return visitor.visitAddBlockHeader(self)
            else:
                return visitor.visitChildren(self)




    def addBlockHeader(self):

        localctx = PatchfileParser.AddBlockHeaderContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_addBlockHeader)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 20
            self.match(PatchfileParser.ADD)
            self.state = 21
            self.match(PatchfileParser.FILENAME)
            self.state = 22
            self.locationToken()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AddBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BEGIN_PATCH(self):
            return self.getToken(PatchfileParser.BEGIN_PATCH, 0)

        def addBlockText(self):
            return self.getTypedRuleContext(PatchfileParser.AddBlockTextContext,0)


        def END_PATCH(self):
            return self.getToken(PatchfileParser.END_PATCH, 0)

        def addBlockHeader(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.AddBlockHeaderContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.AddBlockHeaderContext,i)


        def getRuleIndex(self):
            return PatchfileParser.RULE_addBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAddBlock" ):
                listener.enterAddBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAddBlock" ):
                listener.exitAddBlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAddBlock" ):
                return visitor.visitAddBlock(self)
            else:
                return visitor.visitChildren(self)




    def addBlock(self):

        localctx = PatchfileParser.AddBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_addBlock)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 25 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 24
                self.addBlockHeader()
                self.state = 27 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==1):
                    break

            self.state = 29
            self.match(PatchfileParser.BEGIN_PATCH)
            self.state = 30
            self.addBlockText()
            self.state = 31
            self.match(PatchfileParser.END_PATCH)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AddBlockTextContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def AS_TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(PatchfileParser.AS_TEXT)
            else:
                return self.getToken(PatchfileParser.AS_TEXT, i)

        def getRuleIndex(self):
            return PatchfileParser.RULE_addBlockText

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAddBlockText" ):
                listener.enterAddBlockText(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAddBlockText" ):
                listener.exitAddBlockText(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAddBlockText" ):
                return visitor.visitAddBlockText(self)
            else:
                return visitor.visitChildren(self)




    def addBlockText(self):

        localctx = PatchfileParser.AddBlockTextContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_addBlockText)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 34 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 33
                self.match(PatchfileParser.AS_TEXT)
                self.state = 36 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==18):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RemoveBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REMOVE(self):
            return self.getToken(PatchfileParser.REMOVE, 0)

        def FILENAME(self):
            return self.getToken(PatchfileParser.FILENAME, 0)

        def locationToken(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.LocationTokenContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.LocationTokenContext,i)


        def DASH(self):
            return self.getToken(PatchfileParser.DASH, 0)

        def getRuleIndex(self):
            return PatchfileParser.RULE_removeBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRemoveBlock" ):
                listener.enterRemoveBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRemoveBlock" ):
                listener.exitRemoveBlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRemoveBlock" ):
                return visitor.visitRemoveBlock(self)
            else:
                return visitor.visitChildren(self)




    def removeBlock(self):

        localctx = PatchfileParser.RemoveBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_removeBlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 38
            self.match(PatchfileParser.REMOVE)
            self.state = 39
            self.match(PatchfileParser.FILENAME)
            self.state = 40
            self.locationToken()
            self.state = 41
            self.match(PatchfileParser.DASH)
            self.state = 42
            self.locationToken()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ReplaceNthBlockHeaderContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REPLACE(self):
            return self.getToken(PatchfileParser.REPLACE, 0)

        def FILENAME(self):
            return self.getToken(PatchfileParser.FILENAME, 0)

        def locationToken(self):
            return self.getTypedRuleContext(PatchfileParser.LocationTokenContext,0)


        def getRuleIndex(self):
            return PatchfileParser.RULE_replaceNthBlockHeader

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterReplaceNthBlockHeader" ):
                listener.enterReplaceNthBlockHeader(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitReplaceNthBlockHeader" ):
                listener.exitReplaceNthBlockHeader(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitReplaceNthBlockHeader" ):
                return visitor.visitReplaceNthBlockHeader(self)
            else:
                return visitor.visitChildren(self)




    def replaceNthBlockHeader(self):

        localctx = PatchfileParser.ReplaceNthBlockHeaderContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_replaceNthBlockHeader)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 44
            self.match(PatchfileParser.REPLACE)
            self.state = 45
            self.match(PatchfileParser.FILENAME)
            self.state = 46
            self.locationToken()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ReplaceNthBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BEGIN_CONTENT(self):
            return self.getToken(PatchfileParser.BEGIN_CONTENT, 0)

        def replaceBlockText(self):
            return self.getTypedRuleContext(PatchfileParser.ReplaceBlockTextContext,0)


        def END_CONTENT(self):
            return self.getToken(PatchfileParser.END_CONTENT, 0)

        def BEGIN_PATCH(self):
            return self.getToken(PatchfileParser.BEGIN_PATCH, 0)

        def addBlockText(self):
            return self.getTypedRuleContext(PatchfileParser.AddBlockTextContext,0)


        def END_PATCH(self):
            return self.getToken(PatchfileParser.END_PATCH, 0)

        def replaceNthBlockHeader(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.ReplaceNthBlockHeaderContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.ReplaceNthBlockHeaderContext,i)


        def getRuleIndex(self):
            return PatchfileParser.RULE_replaceNthBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterReplaceNthBlock" ):
                listener.enterReplaceNthBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitReplaceNthBlock" ):
                listener.exitReplaceNthBlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitReplaceNthBlock" ):
                return visitor.visitReplaceNthBlock(self)
            else:
                return visitor.visitChildren(self)




    def replaceNthBlock(self):

        localctx = PatchfileParser.ReplaceNthBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_replaceNthBlock)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 49 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 48
                self.replaceNthBlockHeader()
                self.state = 51 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==3):
                    break

            self.state = 53
            self.match(PatchfileParser.BEGIN_CONTENT)
            self.state = 54
            self.replaceBlockText()
            self.state = 55
            self.match(PatchfileParser.END_CONTENT)
            self.state = 56
            self.match(PatchfileParser.BEGIN_PATCH)
            self.state = 57
            self.addBlockText()
            self.state = 58
            self.match(PatchfileParser.END_PATCH)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ReplaceAllBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REPLACE_ALL(self):
            return self.getToken(PatchfileParser.REPLACE_ALL, 0)

        def FILENAME(self):
            return self.getToken(PatchfileParser.FILENAME, 0)

        def BEGIN_CONTENT(self):
            return self.getToken(PatchfileParser.BEGIN_CONTENT, 0)

        def replaceBlockText(self):
            return self.getTypedRuleContext(PatchfileParser.ReplaceBlockTextContext,0)


        def END_CONTENT(self):
            return self.getToken(PatchfileParser.END_CONTENT, 0)

        def BEGIN_PATCH(self):
            return self.getToken(PatchfileParser.BEGIN_PATCH, 0)

        def addBlockText(self):
            return self.getTypedRuleContext(PatchfileParser.AddBlockTextContext,0)


        def END_PATCH(self):
            return self.getToken(PatchfileParser.END_PATCH, 0)

        def getRuleIndex(self):
            return PatchfileParser.RULE_replaceAllBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterReplaceAllBlock" ):
                listener.enterReplaceAllBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitReplaceAllBlock" ):
                listener.exitReplaceAllBlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitReplaceAllBlock" ):
                return visitor.visitReplaceAllBlock(self)
            else:
                return visitor.visitChildren(self)




    def replaceAllBlock(self):

        localctx = PatchfileParser.ReplaceAllBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_replaceAllBlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 60
            self.match(PatchfileParser.REPLACE_ALL)
            self.state = 61
            self.match(PatchfileParser.FILENAME)
            self.state = 62
            self.match(PatchfileParser.BEGIN_CONTENT)
            self.state = 63
            self.replaceBlockText()
            self.state = 64
            self.match(PatchfileParser.END_CONTENT)
            self.state = 65
            self.match(PatchfileParser.BEGIN_PATCH)
            self.state = 66
            self.addBlockText()
            self.state = 67
            self.match(PatchfileParser.END_PATCH)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ReplaceBlockTextContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def CONTENT_TEXT(self, i:int=None):
            if i is None:
                return self.getTokens(PatchfileParser.CONTENT_TEXT)
            else:
                return self.getToken(PatchfileParser.CONTENT_TEXT, i)

        def getRuleIndex(self):
            return PatchfileParser.RULE_replaceBlockText

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterReplaceBlockText" ):
                listener.enterReplaceBlockText(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitReplaceBlockText" ):
                listener.exitReplaceBlockText(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitReplaceBlockText" ):
                return visitor.visitReplaceBlockText(self)
            else:
                return visitor.visitChildren(self)




    def replaceBlockText(self):

        localctx = PatchfileParser.ReplaceBlockTextContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_replaceBlockText)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 70 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 69
                self.match(PatchfileParser.CONTENT_TEXT)
                self.state = 72 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not (_la==20):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RootContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def addBlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.AddBlockContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.AddBlockContext,i)


        def removeBlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.RemoveBlockContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.RemoveBlockContext,i)


        def replaceNthBlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.ReplaceNthBlockContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.ReplaceNthBlockContext,i)


        def replaceAllBlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(PatchfileParser.ReplaceAllBlockContext)
            else:
                return self.getTypedRuleContext(PatchfileParser.ReplaceAllBlockContext,i)


        def getRuleIndex(self):
            return PatchfileParser.RULE_root

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRoot" ):
                listener.enterRoot(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRoot" ):
                listener.exitRoot(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRoot" ):
                return visitor.visitRoot(self)
            else:
                return visitor.visitChildren(self)




    def root(self):

        localctx = PatchfileParser.RootContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_root)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 80
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & 30) != 0):
                self.state = 78
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [1]:
                    self.state = 74
                    self.addBlock()
                    pass
                elif token in [2]:
                    self.state = 75
                    self.removeBlock()
                    pass
                elif token in [3]:
                    self.state = 76
                    self.replaceNthBlock()
                    pass
                elif token in [4]:
                    self.state = 77
                    self.replaceAllBlock()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 82
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class LocationTokenContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return PatchfileParser.RULE_locationToken

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class FunctionContext(LocationTokenContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a PatchfileParser.LocationTokenContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def FUNCTION(self):
            return self.getToken(PatchfileParser.FUNCTION, 0)
        def FUNCTION_NAME(self):
            return self.getToken(PatchfileParser.FUNCTION_NAME, 0)
        def OPEN_BLOCK(self):
            return self.getToken(PatchfileParser.OPEN_BLOCK, 0)
        def INTEGER(self):
            return self.getToken(PatchfileParser.INTEGER, 0)
        def CLOSE_BLOCK(self):
            return self.getToken(PatchfileParser.CLOSE_BLOCK, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFunction" ):
                listener.enterFunction(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFunction" ):
                listener.exitFunction(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFunction" ):
                return visitor.visitFunction(self)
            else:
                return visitor.visitChildren(self)


    class EndContext(LocationTokenContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a PatchfileParser.LocationTokenContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def END(self):
            return self.getToken(PatchfileParser.END, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnd" ):
                listener.enterEnd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnd" ):
                listener.exitEnd(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitEnd" ):
                return visitor.visitEnd(self)
            else:
                return visitor.visitChildren(self)


    class LineNumberContext(LocationTokenContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a PatchfileParser.LocationTokenContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def INTEGER(self):
            return self.getToken(PatchfileParser.INTEGER, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterLineNumber" ):
                listener.enterLineNumber(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitLineNumber" ):
                listener.exitLineNumber(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLineNumber" ):
                return visitor.visitLineNumber(self)
            else:
                return visitor.visitChildren(self)



    def locationToken(self):

        localctx = PatchfileParser.LocationTokenContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_locationToken)
        self._la = 0 # Token type
        try:
            self.state = 96
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [8, 10]:
                localctx = PatchfileParser.FunctionContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 84
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==10:
                    self.state = 83
                    self.match(PatchfileParser.OPEN_BLOCK)


                self.state = 86
                self.match(PatchfileParser.FUNCTION)
                self.state = 87
                self.match(PatchfileParser.FUNCTION_NAME)
                self.state = 89
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==12:
                    self.state = 88
                    self.match(PatchfileParser.INTEGER)


                self.state = 92
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==11:
                    self.state = 91
                    self.match(PatchfileParser.CLOSE_BLOCK)


                pass
            elif token in [12]:
                localctx = PatchfileParser.LineNumberContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 94
                self.match(PatchfileParser.INTEGER)
                pass
            elif token in [9]:
                localctx = PatchfileParser.EndContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 95
                self.match(PatchfileParser.END)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from __future__ import annotations

from flash_patcher.antlr_source.PatchfileParser import PatchfileParser

from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.inject.location.constant_injection_location import ConstantInjectionLocation
from flash_patcher.inject.location.injection_location import InjectionLocation

class FindContentManager:
    """A class designed to find specified content within a file,
    and replace it with placeholder text.
    """

    context: PatchfileParser.LocationTokenContext = None
    search_content: str

    def __init__(
        self: FindContentManager,
        symbolic_location: PatchfileParser.LocationTokenContext,
        search_content: str,
    ) -> None:
        self.context = symbolic_location
        self.search_content = search_content

    def resolve(
        self: FindContentManager,
        file_content: str,
        exception: ErrorManager,
    ) -> tuple[list[str], InjectionLocation]:
        """Remove the specified text, and return an InjectionLocation that points to
        the location where the new content should be added.

        This will throw an exception if the instance specified is out of bounds
        (for example, requesting the 6th instance when there are only 5 in the file).

        (We don't want to do the full replacement here, because we want to process 
        secondary commands within the patch block with the SingleInjectionManager.)
        """

        instance_number = None

        # Integer context means Nth instance
        if isinstance(self.context, PatchfileParser.LineNumberContext):
            instance_number = int(self.context.INTEGER().getText())
            current_index, occurrences_found = \
                self.find_nth_instance_after(file_content, 0, instance_number)

        # Function context means Nth instance from the start of the function
        # We always take the first function with this name
        elif isinstance(self.context, PatchfileParser.FunctionContext):
            function_name = self.context.FUNCTION_NAME().getText()

            instance_number = 1
            if self.context.INTEGER():
                instance_number = int(self.context.INTEGER().getText())

            function_index_1 = file_content.find(f"function {function_name}")
            function_index_2 = file_content.find(f"{function_name} = function")

            if function_index_1 == -1 and function_index_2 == -1:
                exception.raise_(
                    f"""The specified function ({function_name}) could not be found!!"""
                )

            function_index = min(
                function_index_1 if function_index_1 != -1 else float('inf'),
                function_index_2 if function_index_2 != -1 else float('inf'),
            )

            current_index, occurrences_found = \
                self.find_nth_instance_after(file_content, function_index, instance_number)

        # End context means the last instance
        elif isinstance(self.context, PatchfileParser.EndContext):
            current_index = file_content.rfind(self.search_content)
            instance_number = 0
            occurrences_found = -1 if current_index == -1 else 0

        # If N occurrences are found, get the index of the Nth occurrence
        if occurrences_found == instance_number:
            split_content = [
                file_content[:current_index],
                file_content[current_index + len(self.search_content):],
            ]
            instance_number = 1
        else:
            # If less than N occurrences are found, raise an error
            exception.raise_(
                f"""Could not find ({self.context.getText()})th instance of the content:
                {self.search_content}
                """
            )

        first_section = split_content[:instance_number]
        second_section = split_content[instance_number:]

        location = ConstantInjectionLocation(len(first_section))

        first_section.extend(second_section)

        return first_section, location

    def find_nth_instance_after(
        self: FindContentManager,
        file_content: str,
        start_index: int,
        instance_number: int,
    ) -> tuple(int, int):
        """Find the Nth instance of a string, starting at the provided location.
        Will return a tuple of (Nth instance index in string, num occurrences).
        """
        # Start searching for substring A after the first occurrence of substring B
        current_index = start_index
        occurrences_found = 0

        while occurrences_found < instance_number:
            # Find the next occurrence of substring A
            current_index = file_content.find(self.search_content, current_index)

            # If no more occurrences are found, break out of the loop
            if current_index == -1:
                break

            # Increment the count of occurrences found
            occurrences_found += 1

            # Move the current index to start searching after the current occurrence
            current_index += 1

        return current_index - 1, occurrences_found



# Generated from ../flash_patcher/antlr_source/AssetPack.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .AssetPackParser import AssetPackParser
else:
    from AssetPackParser import AssetPackParser

# This class defines a complete generic visitor for a parse tree produced by AssetPackParser.

class AssetPackVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by AssetPackParser#addAssetBlock.
    def visitAddAssetBlock(self, ctx:AssetPackParser.AddAssetBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AssetPackParser#root.
    def visitRoot(self, ctx:AssetPackParser.RootContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by AssetPackParser#file_name.
    def visitFile_name(self, ctx:AssetPackParser.File_nameContext):
        return self.visitChildren(ctx)



del AssetPackParser


from __future__ import annotations

from pathlib import Path

from flash_patcher.antlr_source.StagefileParser import StagefileParser
from flash_patcher.antlr_source.StagefileVisitor import StagefileVisitor

from flash_patcher.parse.asset import AssetPackManager
from flash_patcher.parse.patch import PatchfileManager

class StagefileProcessor (StagefileVisitor):
    """This class inherits from the ANTLR visitor to process stage files.
    
    It will automatically take in the file syntax tree 
    and perform all injections and asset adds in it.
    """

    # Folder the patch files are in
    folder: Path

    # Decomp location of SWF
    decomp_location: Path

    # Decomp location of SWF scripts
    decomp_location_with_scripts: Path

    # Set of scripts that were modified and need recompilation
    modified_scripts: set[Path]

    def __init__(
        self: StagefileProcessor,
        folder: Path,
        decomp_location: Path,
        decomp_location_with_scripts: Path
     ) -> None:
        self.decomp_location = decomp_location
        self.decomp_location_with_scripts = decomp_location_with_scripts
        self.folder = folder

        self.modified_scripts = set()

    def visitPatchFile(
        self: StagefileProcessor,
        ctx: StagefileParser.PatchFileContext
    ) -> None:
        """When we encounter a patch file, we should open and process it"""
        self.modified_scripts |= PatchfileManager(
            self.decomp_location_with_scripts, self.folder / ctx.getText()
        ).parse()

    def visitAssetPackFile(
        self: StagefileProcessor,
        ctx: StagefileParser.AssetPackFileContext
    ) -> None:
        """When we encounter an asset pack, we should open and process it"""
        self.modified_scripts |= AssetPackManager(
            self.folder, self.decomp_location, self.folder / ctx.getText()
        ).parse()

    def visitRoot(self: StagefileProcessor, ctx: StagefileParser.RootContext) -> set:
        """Root function. Call this when running the visitor."""
        super().visitRoot(ctx)
        return self.modified_scripts



import io
import sys
from logging import error
from typing import Callable

from flash_patcher.exception.dependency import DependencyError

def run_without_antlr_errors(function: Callable[..., any]) -> any:
    """Run a command while suppressing ANTLR version mismatch warnings.
    We also want to raise an exception if there is a parse error, which ANTLR doesn't do.

    ANTLR does not provide an option to suppress these by default, so we need to do it manually.
    """

    # Ignore stdout, ANTLR only uses it for warnings of version mismatch
    sys_stdout_backup = sys.stdout
    sys_stderr_backup = sys.stderr
    sys.stdout = io.StringIO()
    sys.stderr = io.StringIO()

    output = function()

    captured_output = sys.stderr.getvalue()

    # Restore stdout
    sys.stderr = sys_stderr_backup
    sys.stdout = sys_stdout_backup

    process_captured_output(captured_output)

    return output

def process_captured_output(captured_output = str) -> None:
    """Process the output from stderr, and error out if an error occurred."""
    if captured_output != "":
        error_mesg = f"""Processing halted due to lex and parse errors.
            More info:\n{captured_output}"""
        error(error_mesg)
        raise DependencyError(error_mesg)



# Generated from ../flash_patcher/antlr_source/PatchfileParser.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .PatchfileParser import PatchfileParser
else:
    from PatchfileParser import PatchfileParser

# This class defines a complete generic visitor for a parse tree produced by PatchfileParser.

class PatchfileParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by PatchfileParser#addBlockHeader.
    def visitAddBlockHeader(self, ctx:PatchfileParser.AddBlockHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#addBlock.
    def visitAddBlock(self, ctx:PatchfileParser.AddBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#addBlockText.
    def visitAddBlockText(self, ctx:PatchfileParser.AddBlockTextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#removeBlock.
    def visitRemoveBlock(self, ctx:PatchfileParser.RemoveBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#replaceNthBlockHeader.
    def visitReplaceNthBlockHeader(self, ctx:PatchfileParser.ReplaceNthBlockHeaderContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#replaceNthBlock.
    def visitReplaceNthBlock(self, ctx:PatchfileParser.ReplaceNthBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#replaceAllBlock.
    def visitReplaceAllBlock(self, ctx:PatchfileParser.ReplaceAllBlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#replaceBlockText.
    def visitReplaceBlockText(self, ctx:PatchfileParser.ReplaceBlockTextContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#root.
    def visitRoot(self, ctx:PatchfileParser.RootContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#function.
    def visitFunction(self, ctx:PatchfileParser.FunctionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#lineNumber.
    def visitLineNumber(self, ctx:PatchfileParser.LineNumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by PatchfileParser#end.
    def visitEnd(self, ctx:PatchfileParser.EndContext):
        return self.visitChildren(ctx)



del PatchfileParser


from __future__ import annotations

from pathlib import Path

from flash_patcher.antlr_source.AssetPackLexer import AssetPackLexer
from flash_patcher.antlr_source.AssetPackParser import AssetPackParser

from flash_patcher.parse.common import CommonParseManager
from flash_patcher.parse.visitor.asset_visitor import AssetPackProcessor

class AssetPackManager:
    """Manage asset pack files."""
    asset_pack_processor: AssetPackProcessor
    file: Path

    def __init__(
        self: AssetPackManager,
        folder: Path,
        decomp_location: Path,
        file: Path,
    ) -> None:
        self.asset_pack_processor = AssetPackProcessor(folder, decomp_location)
        self.file = file

    def parse(self: AssetPackManager) -> set:
        """Parse a single asset pack.
        
        This class handles everything to do with preprocessing (opening the file, etc.)
        Everything within the file will be handled by the AssetPackProcessor
        """
        asset_pack = CommonParseManager(AssetPackLexer, AssetPackParser).get_root(self.file)

        return self.asset_pack_processor.visitRoot(asset_pack)



# Generated from ../flash_patcher/antlr_source/Stagefile.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,4,18,2,0,7,0,2,1,7,1,2,2,7,2,1,0,1,0,1,1,1,1,1,2,1,2,5,2,13,
        8,2,10,2,12,2,16,9,2,1,2,0,0,3,0,2,4,0,0,16,0,6,1,0,0,0,2,8,1,0,
        0,0,4,14,1,0,0,0,6,7,5,1,0,0,7,1,1,0,0,0,8,9,5,2,0,0,9,3,1,0,0,0,
        10,13,3,0,0,0,11,13,3,2,1,0,12,10,1,0,0,0,12,11,1,0,0,0,13,16,1,
        0,0,0,14,12,1,0,0,0,14,15,1,0,0,0,15,5,1,0,0,0,16,14,1,0,0,0,2,12,
        14
    ]

class StagefileParser ( Parser ):

    grammarFileName = "Stagefile.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [  ]

    symbolicNames = [ "<INVALID>", "PATCH_FILE", "ASSET_PACK_FILE", "WHITESPACE", 
                      "COMMENT" ]

    RULE_patchFile = 0
    RULE_assetPackFile = 1
    RULE_root = 2

    ruleNames =  [ "patchFile", "assetPackFile", "root" ]

    EOF = Token.EOF
    PATCH_FILE=1
    ASSET_PACK_FILE=2
    WHITESPACE=3
    COMMENT=4

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class PatchFileContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def PATCH_FILE(self):
            return self.getToken(StagefileParser.PATCH_FILE, 0)

        def getRuleIndex(self):
            return StagefileParser.RULE_patchFile

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPatchFile" ):
                listener.enterPatchFile(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPatchFile" ):
                listener.exitPatchFile(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPatchFile" ):
                return visitor.visitPatchFile(self)
            else:
                return visitor.visitChildren(self)




    def patchFile(self):

        localctx = StagefileParser.PatchFileContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_patchFile)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 6
            self.match(StagefileParser.PATCH_FILE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AssetPackFileContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ASSET_PACK_FILE(self):
            return self.getToken(StagefileParser.ASSET_PACK_FILE, 0)

        def getRuleIndex(self):
            return StagefileParser.RULE_assetPackFile

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAssetPackFile" ):
                listener.enterAssetPackFile(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAssetPackFile" ):
                listener.exitAssetPackFile(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAssetPackFile" ):
                return visitor.visitAssetPackFile(self)
            else:
                return visitor.visitChildren(self)




    def assetPackFile(self):

        localctx = StagefileParser.AssetPackFileContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_assetPackFile)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.match(StagefileParser.ASSET_PACK_FILE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RootContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def patchFile(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(StagefileParser.PatchFileContext)
            else:
                return self.getTypedRuleContext(StagefileParser.PatchFileContext,i)


        def assetPackFile(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(StagefileParser.AssetPackFileContext)
            else:
                return self.getTypedRuleContext(StagefileParser.AssetPackFileContext,i)


        def getRuleIndex(self):
            return StagefileParser.RULE_root

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRoot" ):
                listener.enterRoot(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRoot" ):
                listener.exitRoot(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRoot" ):
                return visitor.visitRoot(self)
            else:
                return visitor.visitChildren(self)




    def root(self):

        localctx = StagefileParser.RootContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_root)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 14
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1 or _la==2:
                self.state = 12
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [1]:
                    self.state = 10
                    self.patchFile()
                    pass
                elif token in [2]:
                    self.state = 11
                    self.assetPackFile()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 16
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from __future__ import annotations

import base64
from logging import error, info
from pathlib import Path

from flash_patcher.compile.ffdec import FFDecInterface
from flash_patcher.exception.dependency import DependencyError

class CompilationManager:
    """Manage Flash compilation and decompilation, including caching.

    This class should not call FFDec directly, instead it should use the FFDecInterface.
    """

    decompiler: FFDecInterface

    def __init__(self: CompilationManager) -> None:
        self.decompiler = FFDecInterface()

    def decompile(
        self: CompilationManager,
        inputfile: Path,
        drop_cache: bool = False,
        xml_mode: bool = False
    ) -> Path:
        """Decompile the SWF and return the decompilation location.

        This uses caching to save time.

        inputfile: the SWF to decompile
        drop_cache: if True, will force decompilation instead of using cached files
        xml_mode: if True, will dump XML instead of decompiling normally
        """
        # Validity checking: Decompilation path and input file
        if not Path("./.Patcher-Temp").exists():
            Path("./.Patcher-Temp").mkdir()

        if not inputfile.exists():
            failure_mesg = f"""Could not locate the SWF file: {inputfile}.
            Aborting..."""

            error(failure_mesg)
            raise FileNotFoundError(failure_mesg)

        # Decompile swf into temp folder called ./.Patcher-Temp/[swf name, base32 encoded]
        cache_location = Path(
            "./.Patcher-Temp/",
            base64.b32encode(
                bytes(inputfile.name, "utf-8"),
            ).decode("ascii"),
        )

        # If the cache is dropped or nonexistent, rerun decompiler
        if drop_cache or (not Path(cache_location).exists()):
            if not Path(cache_location).exists() and not xml_mode:
                Path(cache_location).mkdir()

            info("Beginning decompilation...")

            decomp = None

            if xml_mode:
                decomp = self.decompiler.dump_xml(inputfile, cache_location)
                info("XML decompilation mode.")
            else:
                decomp = self.decompiler.export_scripts(inputfile, cache_location)

            if not decomp:
                failure_mesg = f"""FFDec couldn't decompile the SWF file: {inputfile}.
                    Aborting..."""

                error(failure_mesg)
                raise DependencyError(failure_mesg)

        else:
            info("Detected cached decompilation. Skipping...")

        return cache_location

    def recompile_with_check(
        self: CompilationManager,
        part: str,
        decomp_location: Path,
        swf: Path,
        output: Path,
    ) -> None:
        """Recompile the SWF part, with a check for program errors."""
        recomp = self.decompiler.recompile_data(part, decomp_location, swf, output)

        if not recomp:
            failure_mesg = f"""FFDec couldn't recompile the SWF file: {swf}.
                Aborting.."""

            error(failure_mesg)
            raise DependencyError(failure_mesg)

    def recompile(
        self: CompilationManager,
        injection: Path,
        inputfile: Path,
        output: Path,
        recompile_all: bool = False,
        xml_mode: bool = False,
    ) -> None:
        """Recompile the SWF after injection is complete.

        inputfile: The base SWF to use for missing files
        outputfile: The location to save the output
        recompile_all: If this is set to False, will only recompile scripts
        """
        if xml_mode:
            self.decompiler.rebuild_xml(injection, output)
            return

        self.recompile_with_check("Script", injection, inputfile, output)

        if recompile_all:
            for part in ("Images", "Sounds", "Shapes", "Text"):
                # FFDec doesn't have a way to import everything at once.
                # We re-import iteratively.
                self.recompile_with_check(part, injection, output, output)



from __future__ import annotations

from flash_patcher.antlr_source.PatchfileParser import PatchfileParser

from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.inject.location.injection_location import InjectionLocation

class ParserInjectionLocation (InjectionLocation):
    """Store the location within a file to inject at.

    This will resolve a symbolic location (like "end") into a line number,
    which can then be injected.

    This uses the parser context to find the correct location.
    """

    context: PatchfileParser.LocationTokenContext = None

    def __init__(
        self: ParserInjectionLocation,
        symbolic_location: PatchfileParser.LocationTokenContext
    ) -> None:
        self.context = symbolic_location

    def resolve(
        self: ParserInjectionLocation,
        file_content: list[str],
        is_add: bool,
        exception: ErrorManager,
    ) -> int | None:
        """Resolve the injection location in the given file, or None if none exists.
        
        If it was unable to resolve, use error_line_no to throw and exception.
        """
        line_no = None
        if isinstance(self.context, PatchfileParser.LineNumberContext):
            line_no = self.resolve_line_no(file_content, is_add, exception)

        elif isinstance(self.context, PatchfileParser.FunctionContext):
            line_no = self.resolve_function(file_content, exception)

        elif isinstance(self.context, PatchfileParser.EndContext):
            line_no = self.resolve_end(file_content)

        # Unknown injection location
        else:
            exception.context = self.context
            exception.raise_(
                f"""Invalid file location.
                Expected keyword or integer (got type "{type(self.context)}").""",
            )

        if line_no is None:
            return line_no

        return max(line_no, 0)

    def resolve_line_no(
        self: ParserInjectionLocation,
        file_content: list[str],
        is_add: bool,
        exception: ErrorManager
    ) -> int:
        """Resolve the injection location if it's a line number."""
        line_no = int(self.context.INTEGER().getText())

        # ensure we inject before the line
        if is_add:
            line_no -= 1

        self.verify_line_no(line_no, file_content, exception)

        return line_no

    def resolve_function(
        self: ParserInjectionLocation,
        file_content: list[str],
        exception: ErrorManager,
    ) -> int | None:
        """Resolve the injection location if it's a function + offset."""
        line_no = None
        for i, line in enumerate(file_content):
            # Do some replacement to ensure that function names are processed correctly
            line = line.replace("(", " ")
            line = line.replace(")", " ")

            if line.split()[:2] == ["function", self.context.FUNCTION_NAME().getText()] \
                or line.split()[:3] == [self.context.FUNCTION_NAME().getText(), "=", "function"]:
                # We need to add after the specified line
                line_no = i + 1
                break

        if line_no is not None:
            if self.context.INTEGER():
                line_no += int(self.context.INTEGER().getText())

            self.verify_line_no(line_no, file_content, exception)

        return line_no

    def resolve_end(
        self: ParserInjectionLocation,
        file_content: list[str],
    ) -> int:
        """Resolve the injection location if it's the end of the file."""
        return len(file_content)

    def verify_line_no(
        self: ParserInjectionLocation,
        line_no: int,
        file_content: list[str],
        exception: ErrorManager,
    ) -> None:
        """Verify that the line number is not beyond EOF, and throw an exception if it is."""
        if line_no > len(file_content):
            exception.raise_(
                f"""Out of bounds file location {line_no}.
                The provided location is outside the maximum line number in the file."""
            )



#!/usr/bin/env python3

import argparse
from pathlib import Path

from flash_patcher.patcher import main

def cli():
    """Run Flash Patcher from the CLI."""
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--inputswf",
        dest="input_swf",
        type=str,
        required=True,
        help="Input SWF file",
    )

    parser.add_argument(
        "--folder",
        dest="folder",
        type=str,
        required=True,
        help="Folder with patch files",
    )

    parser.add_argument(
        "--stagefile",
        dest="stage_file",
        type=str,
        required=True,
        help="Stage file name",
    )

    parser.add_argument(
        "--outputswf",
        dest="output_swf",
        type=str,
        required=True,
        help="Output SWF file",
    )

    parser.add_argument(
        "--invalidateCache",
        dest="drop_cache",
        default=False,
        action="store_true",
        help="Invalidate cached decompilation files",
    )

    parser.add_argument(
        "--all",
        dest="recompile_all",
        default=False,
        action="store_true",
        help="Recompile the whole SWF (if this is off, only scripts will recompile)",
    )

    parser.add_argument(
        "--xml",
        dest="xml_mode",
        default=False,
        action="store_true",
        help="Inject into an XML decompilation instead of standard syntax",
    )

    args = parser.parse_args()

    main(
        Path(args.input_swf),
        Path(args.folder),
        Path(args.stage_file),
        Path(args.output_swf),
        drop_cache=args.drop_cache,
        recompile_all=args.recompile_all,
        xml_mode=args.xml_mode,
    )


if __name__ == "__main__":
    cli()



from __future__ import annotations

from logging import error

from flash_patcher.exception.injection import InjectionError

class ErrorManager:
    """Handle exceptions thrown by Flash Patcher during the injection stage."""
    patch_file: str
    line_no: int
    context: str

    def __init__(
        self: ErrorManager,
        patch_file: str,
        line_no: int,
        context: str | None = None,
    ) -> None:
        self.patch_file = patch_file
        self.line_no = line_no

        # extra info is usually the content of the offending line
        self.context = context

    def raise_(self: ErrorManager, mesg: str) -> None:
        """Raise an InjectionError with the requested information."""
        error_mesg = f"""InjectionError at {self.patch_file}, line {self.line_no}
            Additional context: {self.context}
            {mesg}
            Aborting..."""

        error(error_mesg)
        raise InjectionError(mesg)



# Generated from ../flash_patcher/antlr_source/AssetPack.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .AssetPackParser import AssetPackParser
else:
    from AssetPackParser import AssetPackParser

# This class defines a complete listener for a parse tree produced by AssetPackParser.
class AssetPackListener(ParseTreeListener):

    # Enter a parse tree produced by AssetPackParser#addAssetBlock.
    def enterAddAssetBlock(self, ctx:AssetPackParser.AddAssetBlockContext):
        pass

    # Exit a parse tree produced by AssetPackParser#addAssetBlock.
    def exitAddAssetBlock(self, ctx:AssetPackParser.AddAssetBlockContext):
        pass


    # Enter a parse tree produced by AssetPackParser#root.
    def enterRoot(self, ctx:AssetPackParser.RootContext):
        pass

    # Exit a parse tree produced by AssetPackParser#root.
    def exitRoot(self, ctx:AssetPackParser.RootContext):
        pass


    # Enter a parse tree produced by AssetPackParser#file_name.
    def enterFile_name(self, ctx:AssetPackParser.File_nameContext):
        pass

    # Exit a parse tree produced by AssetPackParser#file_name.
    def exitFile_name(self, ctx:AssetPackParser.File_nameContext):
        pass



del AssetPackParser


from __future__ import annotations

from pathlib import Path

from flash_patcher.antlr_source.PatchfileParser import PatchfileParser
from flash_patcher.antlr_source.PatchfileParserVisitor import PatchfileParserVisitor

from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.inject.bulk_injection import BulkInjectionManager
from flash_patcher.inject.location.parser_injection_location import ParserInjectionLocation
from flash_patcher.inject.find_content import FindContentManager
from flash_patcher.inject.single_injection import SingleInjectionManager
from flash_patcher.util.file_io import FileWritebackManager, read_safe

class PatchfileProcessor (PatchfileParserVisitor):
    """This class inherits from the ANTLR visitor to process patch files.
    
    It will automatically take in the file syntax tree and perform the injections in it. 
    """

    injector: BulkInjectionManager
    modified_scripts: set[Path]
    patch_file_name: Path
    decomp_location_with_scripts: Path

    def __init__(
        self: PatchfileProcessor,
        patch_file_name: Path,
        decomp_location_with_scripts: Path
    ) -> None:
        self.patch_file_name = patch_file_name
        self.decomp_location_with_scripts = decomp_location_with_scripts

        self.injector = BulkInjectionManager()
        self.modified_scripts = set()

    def visitAddBlockHeader(
        self: PatchfileProcessor,
        ctx: PatchfileParser.AddBlockHeaderContext
    ) -> None:
        """Add the headers to the injector metadata"""
        full_path = self.decomp_location_with_scripts / ctx.FILENAME().getText()

        inject_location = ParserInjectionLocation(ctx.locationToken())

        self.injector.add_injection_target(
            SingleInjectionManager(full_path, inject_location, self.patch_file_name, ctx.start.line)
        )
        self.modified_scripts.add(full_path)

    def visitAddBlock(
        self: PatchfileProcessor,
        ctx: PatchfileParser.AddBlockContext
    ) -> None:
        """When we visit an add block, use an injector to manage injection"""
        for header in ctx.addBlockHeader():
            self.visitAddBlockHeader(header)

        stripped_text = ctx.addBlockText().getText()

        if stripped_text[0] == "\n":
            stripped_text = stripped_text[1:]

        self.injector.inject(stripped_text)
        self.injector.clear()

    def visitRemoveBlock(
        self: PatchfileProcessor,
        ctx: PatchfileParser.RemoveBlockContext
    ) -> None:
        """Remove is processed manually as the command is less complex than add."""
        full_path = self.decomp_location_with_scripts / ctx.FILENAME().getText()

        # Open file, delete lines, and close it
        error_manager = ErrorManager(self.patch_file_name, ctx.start.line)

        with FileWritebackManager(full_path, error_manager, readlines=True) as current_file:
            line_start = ParserInjectionLocation(ctx.locationToken(0)) \
                .resolve(current_file, False, error_manager)

            line_end = ParserInjectionLocation(ctx.locationToken(1)) \
                .resolve(current_file, False, error_manager)

            if line_start is None or line_end is None:
                error_manager.raise_(
                    """Could not resolve line start or end.
                    You must provide a valid and in-bounds line number for remove.
                    """
                )

            # Exceptions will be thrown in InjectionLocation if this location is invalid
            for _ in range(line_start, line_end + 1):
                del current_file[line_start - 1]

        self.modified_scripts.add(full_path)

    def visitReplaceNthBlock(
        self: PatchfileProcessor,
        ctx: PatchfileParser.ReplaceNthBlockContext
    ) -> None:
        """Replace the nth block. 
        Find its location as an InjectionLocation, 
        then remove it and perform a standard add-injection at that location.
        """
        for i in ctx.replaceNthBlockHeader():
            full_path = self.decomp_location_with_scripts / i.FILENAME().getText()
            error_manager = ErrorManager(self.patch_file_name, i.start.line)

            current_file = read_safe(full_path, error_manager)
            updated_file, replace_location = FindContentManager(
                i.locationToken(), ctx.replaceBlockText().getText().strip()
            ).resolve(current_file, error_manager)

            injector = SingleInjectionManager(
                full_path, replace_location, full_path, i.start.line
            )

            injector.file_content = updated_file
            injector.inject(ctx.addBlockText().getText().strip(), i.start.line)

            self.modified_scripts.add(full_path)

    def visitRoot(self: PatchfileProcessor, ctx: PatchfileParser.RootContext) -> set:
        """Root function. Call this when running the visitor."""
        super().visitRoot(ctx)
        return self.modified_scripts



from __future__ import annotations

from pathlib import Path

from flash_patcher.antlr_source.StagefileLexer import StagefileLexer
from flash_patcher.antlr_source.StagefileParser import StagefileParser

from flash_patcher.parse.common import CommonParseManager
from flash_patcher.parse.visitor.stage_visitor import StagefileProcessor

class StagefileManager:
    """Manage stage files."""
    stagefile_processor: StagefileProcessor
    stage_file: Path

    def __init__(
        self: StagefileManager,
        folder: Path,
        file: Path,
        decomp_location: Path,
        decomp_location_with_scripts: Path,
    ) -> None:
        self.stagefile_processor = StagefileProcessor(
            folder, decomp_location, decomp_location_with_scripts
        )

        self.stage_file = folder / file

    def parse(self: StagefileManager) -> set:
        """Parse a single stagefile.
        
        This class handles everything to do with preprocessing (opening the file, etc.)
        Everything within the file will be handled by the StagefileProcessor
        """

        stagefile = CommonParseManager(StagefileLexer, StagefileParser).get_root(self.stage_file)

        return self.stagefile_processor.visitRoot(stagefile)



from __future__ import annotations

from flash_patcher.inject.single_injection import SingleInjectionManager

class BulkInjectionManager:
    """Handles injection of code into script files.
    
    This class will take in a list of injections and execute them.
    It should be used to parse blocks of multiple add statements, 
    as it only supports one file position per output file.

    Its contents are built as the input is parsed.
    This will only handle scripts.
    """

    injectors: list[SingleInjectionManager]
    starting_line_no: int

    def __init__(self: BulkInjectionManager) -> None:
        self.injectors = []         # List of single injectors that are primed
        self.injected_content = []   # List of lines to inject
        self.starting_line_no = -1    # The starting line (in the patchfile) of the injection

    # Return the file name of the script being modified
    def add_injection_target(
        self: BulkInjectionManager,
        target: SingleInjectionManager,
    ) -> None:
        """Add an injection target to this injector."""
        self.injectors.append(target)

    def inject(self: BulkInjectionManager, content: str) -> None:
        """Perform loaded injections."""
        for injector in self.injectors:
            injector.inject(content.splitlines(keepends=True), self.starting_line_no)

    def clear(self: BulkInjectionManager) -> None:
        """Clear all loaded content within this injector, to allow it to be used again."""
        self.injectors = []
        self.starting_line_no = -1



# Generated from ../flash_patcher/antlr_source/PatchfileParser.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .PatchfileParser import PatchfileParser
else:
    from PatchfileParser import PatchfileParser

# This class defines a complete listener for a parse tree produced by PatchfileParser.
class PatchfileParserListener(ParseTreeListener):

    # Enter a parse tree produced by PatchfileParser#addBlockHeader.
    def enterAddBlockHeader(self, ctx:PatchfileParser.AddBlockHeaderContext):
        pass

    # Exit a parse tree produced by PatchfileParser#addBlockHeader.
    def exitAddBlockHeader(self, ctx:PatchfileParser.AddBlockHeaderContext):
        pass


    # Enter a parse tree produced by PatchfileParser#addBlock.
    def enterAddBlock(self, ctx:PatchfileParser.AddBlockContext):
        pass

    # Exit a parse tree produced by PatchfileParser#addBlock.
    def exitAddBlock(self, ctx:PatchfileParser.AddBlockContext):
        pass


    # Enter a parse tree produced by PatchfileParser#addBlockText.
    def enterAddBlockText(self, ctx:PatchfileParser.AddBlockTextContext):
        pass

    # Exit a parse tree produced by PatchfileParser#addBlockText.
    def exitAddBlockText(self, ctx:PatchfileParser.AddBlockTextContext):
        pass


    # Enter a parse tree produced by PatchfileParser#removeBlock.
    def enterRemoveBlock(self, ctx:PatchfileParser.RemoveBlockContext):
        pass

    # Exit a parse tree produced by PatchfileParser#removeBlock.
    def exitRemoveBlock(self, ctx:PatchfileParser.RemoveBlockContext):
        pass


    # Enter a parse tree produced by PatchfileParser#replaceNthBlockHeader.
    def enterReplaceNthBlockHeader(self, ctx:PatchfileParser.ReplaceNthBlockHeaderContext):
        pass

    # Exit a parse tree produced by PatchfileParser#replaceNthBlockHeader.
    def exitReplaceNthBlockHeader(self, ctx:PatchfileParser.ReplaceNthBlockHeaderContext):
        pass


    # Enter a parse tree produced by PatchfileParser#replaceNthBlock.
    def enterReplaceNthBlock(self, ctx:PatchfileParser.ReplaceNthBlockContext):
        pass

    # Exit a parse tree produced by PatchfileParser#replaceNthBlock.
    def exitReplaceNthBlock(self, ctx:PatchfileParser.ReplaceNthBlockContext):
        pass


    # Enter a parse tree produced by PatchfileParser#replaceAllBlock.
    def enterReplaceAllBlock(self, ctx:PatchfileParser.ReplaceAllBlockContext):
        pass

    # Exit a parse tree produced by PatchfileParser#replaceAllBlock.
    def exitReplaceAllBlock(self, ctx:PatchfileParser.ReplaceAllBlockContext):
        pass


    # Enter a parse tree produced by PatchfileParser#replaceBlockText.
    def enterReplaceBlockText(self, ctx:PatchfileParser.ReplaceBlockTextContext):
        pass

    # Exit a parse tree produced by PatchfileParser#replaceBlockText.
    def exitReplaceBlockText(self, ctx:PatchfileParser.ReplaceBlockTextContext):
        pass


    # Enter a parse tree produced by PatchfileParser#root.
    def enterRoot(self, ctx:PatchfileParser.RootContext):
        pass

    # Exit a parse tree produced by PatchfileParser#root.
    def exitRoot(self, ctx:PatchfileParser.RootContext):
        pass


    # Enter a parse tree produced by PatchfileParser#function.
    def enterFunction(self, ctx:PatchfileParser.FunctionContext):
        pass

    # Exit a parse tree produced by PatchfileParser#function.
    def exitFunction(self, ctx:PatchfileParser.FunctionContext):
        pass


    # Enter a parse tree produced by PatchfileParser#lineNumber.
    def enterLineNumber(self, ctx:PatchfileParser.LineNumberContext):
        pass

    # Exit a parse tree produced by PatchfileParser#lineNumber.
    def exitLineNumber(self, ctx:PatchfileParser.LineNumberContext):
        pass


    # Enter a parse tree produced by PatchfileParser#end.
    def enterEnd(self, ctx:PatchfileParser.EndContext):
        pass

    # Exit a parse tree produced by PatchfileParser#end.
    def exitEnd(self, ctx:PatchfileParser.EndContext):
        pass



del PatchfileParser


from __future__ import annotations

from abc import ABC, abstractmethod

from flash_patcher.exception.error_manager import ErrorManager

class InjectionLocation (ABC):
    """Store the location within a file to inject at.

    This will resolve into a line number, which can then be injected.
    """

    @abstractmethod
    def resolve(
        self: InjectionLocation,
        file_content: list[str],
        is_add: bool,
        exception: ErrorManager,
    ) -> int | None:
        """Resolve the injection location in the given file, or None if none exists.
        
        If it was unable to resolve, use error_line_no to throw and exception.
        """
        return None



from __future__ import annotations

from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.inject.location.injection_location import InjectionLocation

class ConstantInjectionLocation (InjectionLocation):
    """Store the location within a file to inject at.

    This class stores a constant integer location to inject at.
    """

    location: int

    def __init__(self: ConstantInjectionLocation, location: int) -> None:
        self.location = location

    # pylint: disable=unused-argument
    # This function doesn't use the args, but they're still required
    # due to this being an overload of InjectionLocation.resolve
    def resolve(
        self: ConstantInjectionLocation,
        file_content: list[str],
        is_add: bool,
        exception: ErrorManager,
    ) -> int | None:
        """Return the stored injection location."""
        return self.location



# Generated from ../flash_patcher/antlr_source/Stagefile.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .StagefileParser import StagefileParser
else:
    from StagefileParser import StagefileParser

# This class defines a complete listener for a parse tree produced by StagefileParser.
class StagefileListener(ParseTreeListener):

    # Enter a parse tree produced by StagefileParser#patchFile.
    def enterPatchFile(self, ctx:StagefileParser.PatchFileContext):
        pass

    # Exit a parse tree produced by StagefileParser#patchFile.
    def exitPatchFile(self, ctx:StagefileParser.PatchFileContext):
        pass


    # Enter a parse tree produced by StagefileParser#assetPackFile.
    def enterAssetPackFile(self, ctx:StagefileParser.AssetPackFileContext):
        pass

    # Exit a parse tree produced by StagefileParser#assetPackFile.
    def exitAssetPackFile(self, ctx:StagefileParser.AssetPackFileContext):
        pass


    # Enter a parse tree produced by StagefileParser#root.
    def enterRoot(self, ctx:StagefileParser.RootContext):
        pass

    # Exit a parse tree produced by StagefileParser#root.
    def exitRoot(self, ctx:StagefileParser.RootContext):
        pass



del StagefileParser


from __future__ import annotations

class DependencyError (Exception):
    """Exception raised for dependency-related errors."""

    def __init__(self: DependencyError, message: str) -> None:
        super().__init__(message)



from __future__ import annotations

from pathlib import Path

from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.inject.location.injection_location import InjectionLocation
from flash_patcher.util.file_io import readlines_safe, writelines_safe

class SingleInjectionManager:
    """A position in a named file."""

    # file to inject into
    file_name: Path

    # location to inject at
    file_location: InjectionLocation

    # name of the patch file
    patch_file: Path

    # line number within the pacth file
    patch_line_no: int

    error_manager: ErrorManager
    file_content: list[str]

    def __init__(
        self: SingleInjectionManager,
        file_name: Path,
        file_location: InjectionLocation,
        patch_file: Path,
        patch_line_no: int,
    ) -> None:
        self.file_name = file_name
        self.file_location = file_location
        self.file_content = None

        self.patch_file = patch_file
        self.patch_line_no = patch_line_no

        self.error_manager = ErrorManager(self.patch_file.as_posix(), 0, None)

    def inject(self: SingleInjectionManager, content: list[str], patch_file_line: int) -> None:
        """Inject the content into the file."""
        patch_line_no = patch_file_line

        if self.file_content is None:
            self.file_content = readlines_safe(self.file_name, self.error_manager)

        file_line_no = self.file_location.resolve(self.file_content, True, self.error_manager)

        if file_line_no is None:
            return

        for line in content:
            line_stripped = line.strip("\n\r ")

            # setup error information for the current line
            self.error_manager.patch_file = self.patch_file
            self.error_manager.line_no = patch_line_no
            self.error_manager.context = line_stripped

            # Handle internal commands
            file_line_no, was_secondary = self.handle_secondary_command(
                file_line_no, line_stripped
            )

            # No internal command, just a normal line
            if not was_secondary:
                self.file_content.insert(file_line_no, line)
                file_line_no += 1

            patch_line_no += 1

        writelines_safe(self.file_name, self.file_content)

    def handle_secondary_command(
        self: SingleInjectionManager,
        old_line_no: int,
        command: str,
    ) -> (int, bool):
        """Handle a secondary command within the patch file itself.
        
        For example:
        // cmd: skip 10
        This takes in the current line no. we're injecting at and the command.
        Returns the new line no. after the command, 
        and a boolean on whether any command was executed.
        """
        command_split = command.split()
        match command_split[:3]:
            case ["//", "cmd:", "skip"]:
                # command found
                new_line_no = self.handle_secondary_skip_command(old_line_no, command)
                return new_line_no, True

            case ["//", "cmd:", _]:
                self.error_manager.raise_(
                    """Invalid secondary command.
                    Expected one of: [skip].""",
                )

        return old_line_no, False

    def handle_secondary_skip_command(
        self: SingleInjectionManager,
        old_line_no: int,
        command: str,
    ) -> int:
        """Handle a secondary skip command of the form: // cmd: skip 10

        This will skip 10 lines ahead within the file.
        Returns True if the command is a valid skip command, and updates the line no. accordingly.
        """

        split_line = command.split()
        n_str = split_line[3]

        try:
            skip_amount = int(n_str)
            file_line_no = old_line_no + skip_amount
            return file_line_no

        except ValueError:
            self.error_manager.raise_(
                """Invalid skip amount.
                Expected integer.""",
            )



from __future__ import annotations

from logging import exception, info
from pathlib import Path
from typing import Type

from antlr4 import CommonTokenStream, InputStream
from antlr4.Lexer import Lexer
from antlr4.Parser import Parser, ParserRuleContext

from flash_patcher.exception.dependency import DependencyError
from flash_patcher.exception.error_suppression import run_without_antlr_errors
from flash_patcher.exception.error_manager import ErrorManager
from flash_patcher.util.file_io import read_safe

class CommonParseManager:
    """Common logic for parsing any file type."""

    lexer: Type[Lexer]
    parser: Type[Parser]

    def __init__(
        self: CommonParseManager,
        lexer: Type[Lexer],
        parser: Type[Parser],
    ) -> None:
        self.lexer = lexer
        self.parser = parser

    def parse_input(
        self: CommonParseManager,
        file_content: str,
    ) -> Parser:
        """Subfunction to generate a parser from file content.
        
        This needs to be a separate function to enable separate handling of ANTLR errors in it.
        """
        file_lexer = self.lexer(InputStream(file_content))
        return self.parser(CommonTokenStream(file_lexer))

    def get_root(
        self: CommonParseManager,
        file: Path,
    ) -> ParserRuleContext:
        """Get the root node of the syntax tree from the given file.
        
        This will automatically process all errors.
        Note that the syntax tree root within your grammar MUST have the name `root`.
        """
        info("Processing file: %s", file.as_posix())
        error_manager = ErrorManager(file.as_posix(), 0)
        file_content = read_safe(file, error_manager)

        parser = run_without_antlr_errors(
            lambda: self.parse_input(file_content)
        )

        try:
            tree = run_without_antlr_errors(parser.root)

        # pylint: disable=broad-exception-caught
        # We want to ignore broad exception, as this should handle ANY exception thrown by ANTLR
        except Exception as exc:
            error_mesg = f"""Error while parsing file {file.as_posix()}.
            There is likely additional logging output above.
            """

            exception(error_mesg)
            raise DependencyError(error_mesg) from exc

        return tree
        


from __future__ import annotations

from logging import exception
from pathlib import Path
from typing import Any, Optional, Type

from flash_patcher.exception.error_manager import ErrorManager

class FileWritebackManager:
    """Handle files. Will open and read the content, and writeback when the file is closed."""

    file_location: Path
    error_manager: ErrorManager

    # If set to true, return a list of lines instead of the file as a string
    readlines: bool

    content: str | list[str]

    def __init__(
        self: FileWritebackManager,
        file_location: Path,
        error_manager: ErrorManager,
        readlines: bool = False,
    ) -> None:
        self.file_location = file_location
        self.error_manager = error_manager

        self.readlines = readlines

    def __enter__(self: FileWritebackManager) -> str | list[str]:
        if self.readlines:
            self.content = readlines_safe(self.file_location, self.error_manager)
        else:
            self.content = read_safe(self.file_location, self.error_manager)

        return self.content

    def __exit__(
        self: FileWritebackManager,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[Any],
    ) -> None:
        if isinstance(self.content, str):
            self.content = [self.content]

        writelines_safe(self.file_location, self.content)

def read_safe(file_location: Path, error_manager: ErrorManager) -> str:
    """Read the full content from a file.
    
    Returns a string containing the entire file.
    """
    try:
        with file_location.open() as file:
            return file.read()
    except (FileNotFoundError, IsADirectoryError):
        error_manager.context = file_location.as_posix()
        error_manager.raise_(
            """Invalid injection location.
            Could not find or load SWF decompiled file."""
        )

def readlines_safe(file_location: Path, error_manager: ErrorManager) -> list:
    """Read all lines from a file.
    
    Returns a list, with one entry for each line.
    """
    try:
        with file_location.open() as file:
            return file.readlines()
    except (FileNotFoundError, IsADirectoryError):
        error_manager.context = file_location.as_posix()
        error_manager.raise_(
            """Invalid injection location.
            Could not find or load SWF decompiled file."""
        )

def writelines_safe(path: Path, lines: list[str]) -> None:
    """Write a list of lines to a file."""
    try:
        with path.open("w") as file:
            file.writelines(lines)
    except (FileNotFoundError, IsADirectoryError) as exc:
        mesg = """The provided decompilation is not in a writable location.
            Please ensure you have write access in the current directory."""
        exception(mesg)
        raise exc



# Generated from ../flash_patcher/antlr_source/Stagefile.g4 by ANTLR 4.13.1
from antlr4 import *
if "." in __name__:
    from .StagefileParser import StagefileParser
else:
    from StagefileParser import StagefileParser

# This class defines a complete generic visitor for a parse tree produced by StagefileParser.

class StagefileVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by StagefileParser#patchFile.
    def visitPatchFile(self, ctx:StagefileParser.PatchFileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by StagefileParser#assetPackFile.
    def visitAssetPackFile(self, ctx:StagefileParser.AssetPackFileContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by StagefileParser#root.
    def visitRoot(self, ctx:StagefileParser.RootContext):
        return self.visitChildren(ctx)



del StagefileParser


from __future__ import annotations

import shutil
from logging import exception
from pathlib import Path

from flash_patcher.antlr_source.AssetPackParser import AssetPackParser
from flash_patcher.antlr_source.AssetPackVisitor import AssetPackVisitor

class AssetPackProcessor (AssetPackVisitor):
    """This class inherits from the ANTLR visitor to process asset pack files.
    
    It will automatically take in the file syntax tree and add the files to the SWF.
    """

    asset_folder: Path
    decomp_location: Path
    modified_scripts: set[Path]

    def __init__(self: AssetPackProcessor, asset_folder: Path, decomp_location: Path) -> None:
        self.asset_folder = asset_folder
        self.decomp_location = decomp_location
        self.modified_scripts = set()

    def visitAddAssetBlock(
        self: AssetPackProcessor,
        ctx: AssetPackParser.AddAssetBlockContext
    ) -> None:
        """In an Add Asset block, we should take the specified assets and copy them into the SWF"""
        local_name = ctx.local.getText()
        remote_name = ctx.swf.getText()

        if not Path(self.asset_folder / local_name).exists():
            error_mesg = f"""Could not find asset: {local_name}
            Aborting..."""
            exception(error_mesg)
            raise FileNotFoundError(error_mesg)

        # Create folder and copy things over
        remote_folder = remote_name.split("/")[0]

        if not (self.decomp_location / remote_folder).exists():
            Path.mkdir(self.decomp_location / remote_folder)

        shutil.copyfile(self.asset_folder / local_name, self.decomp_location / remote_name)

        self.modified_scripts.add(self.decomp_location / remote_name)

    def visitRoot(self: AssetPackProcessor, ctx: AssetPackParser.RootContext) -> set:
        """Root function. Call this when running the visitor."""
        super().visitRoot(ctx)
        return self.modified_scripts



from __future__ import annotations

from pathlib import Path

from flash_patcher.antlr_source.PatchfileLexer import PatchfileLexer
from flash_patcher.antlr_source.PatchfileParser import PatchfileParser

from flash_patcher.parse.common import CommonParseManager
from flash_patcher.parse.visitor.patch_visitor import PatchfileProcessor

class PatchfileManager:
    """Manage patch files."""

    file: Path
    patchfile_processor: PatchfileProcessor

    def __init__(self: PatchfileManager, decomp_location: Path, file: Path) -> None:
        self.file = file
        self.patchfile_processor = PatchfileProcessor(self.file, decomp_location)

    def parse(self: PatchfileManager) -> set:
        """Parse a single patch file.

        folder_location: The location of the decompiled scripts from the SWF.
        file_location: The location of the patch file to apply.
        
        This class handles everything to do with preprocessing (opening the file, etc.)
        Everything within the file will be handled by the PatchfileProcessor
        Return the set of modified scripts.
        """
        patchfile = CommonParseManager(PatchfileLexer, PatchfileParser).get_root(self.file)

        return self.patchfile_processor.visitRoot(patchfile)



from logging import basicConfig, exception, info
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

from flash_patcher.compile.compilation import CompilationManager
from flash_patcher.compile.locate_decomp import get_decomp_locations
from flash_patcher.exception.dependency import DependencyError
from flash_patcher.parse.stage import StagefileManager
from flash_patcher.util.file_copy import clean_scripts, copy_file

# pylint: disable=pointless-string-statement
"""
rayyaw's SWF patcher - a tool to patch content into SWF files.

Development: rayyaw, qtkito, GTcreyon
Windows path fix: Jhynjhiruu

Download and updates: https://github.com/rayyaw/flash-patcher

License: CC-BY SA 4.0 (https://creativecommons.org/licenses/by-sa/4.0)

Dependencies: Python 3, JPEXS Decompiler (https://github.com/jindrapetrik/jpexs-decompiler/releases)

Inject arbitrary code, images, and more into existing SWFs!
See the README for documentation and license.
"""

basicConfig(level=1, format="%(levelname)s: %(message)s")

def main(
    inputfile: Path,
    folder: Path,
    stagefile: Path,
    output: Path,
    drop_cache: bool = False,
    recompile_all: bool = False,
    xml_mode: bool = False,
) -> None:
    """Run the patcher."""
    try:
        __version__ = version("flash_patcher")
    except PackageNotFoundError:
        __version__ = "Unit Tests"

    info(f"rayyaw's SWF Patcher - v{__version__}")

    try:
        compiler = CompilationManager()
    except ModuleNotFoundError as exc:
        error_mesg = "Could not locate required dependency: JPEXS Flash Decompiler. Aborting..."
        exception(error_mesg)
        raise DependencyError(error_mesg) from exc

    decomp_location, decomp_location_with_scripts = get_decomp_locations(xml_mode)

    cache_location = compiler.decompile(
        inputfile,
        drop_cache=drop_cache,
        xml_mode=xml_mode,
    )

    # Copy the cache to a different location so we can reuse it
    copy_file(cache_location, decomp_location)

    info("Decompilation finished. Beginning injection...")

    modified_scripts = StagefileManager(
        folder,
        stagefile,
        decomp_location,
        decomp_location_with_scripts
    ).parse()

    info("Injection complete, cleaning up...")

    clean_scripts(decomp_location, modified_scripts)

    info("Recompiling...")

    compiler.recompile(
        decomp_location,
        inputfile,
        output,
        recompile_all=recompile_all,
        xml_mode=xml_mode,
    )

    info("Done.")



# Generated from ../flash_patcher/antlr_source/AssetPack.g4 by ANTLR 4.13.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO

def serializedATN():
    return [
        4,1,4,19,2,0,7,0,2,1,7,1,2,2,7,2,1,0,1,0,1,0,1,0,1,1,5,1,12,8,1,
        10,1,12,1,15,9,1,1,2,1,2,1,2,0,0,3,0,2,4,0,0,16,0,6,1,0,0,0,2,13,
        1,0,0,0,4,16,1,0,0,0,6,7,5,1,0,0,7,8,3,4,2,0,8,9,3,4,2,0,9,1,1,0,
        0,0,10,12,3,0,0,0,11,10,1,0,0,0,12,15,1,0,0,0,13,11,1,0,0,0,13,14,
        1,0,0,0,14,3,1,0,0,0,15,13,1,0,0,0,16,17,5,2,0,0,17,5,1,0,0,0,1,
        13
    ]

class AssetPackParser ( Parser ):

    grammarFileName = "AssetPack.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [  ]

    symbolicNames = [ "<INVALID>", "ADD_ASSET", "FILE_NAME", "WHITESPACE", 
                      "COMMENT" ]

    RULE_addAssetBlock = 0
    RULE_root = 1
    RULE_file_name = 2

    ruleNames =  [ "addAssetBlock", "root", "file_name" ]

    EOF = Token.EOF
    ADD_ASSET=1
    FILE_NAME=2
    WHITESPACE=3
    COMMENT=4

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.13.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class AddAssetBlockContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.local = None # File_nameContext
            self.swf = None # File_nameContext

        def ADD_ASSET(self):
            return self.getToken(AssetPackParser.ADD_ASSET, 0)

        def file_name(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(AssetPackParser.File_nameContext)
            else:
                return self.getTypedRuleContext(AssetPackParser.File_nameContext,i)


        def getRuleIndex(self):
            return AssetPackParser.RULE_addAssetBlock

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterAddAssetBlock" ):
                listener.enterAddAssetBlock(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitAddAssetBlock" ):
                listener.exitAddAssetBlock(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAddAssetBlock" ):
                return visitor.visitAddAssetBlock(self)
            else:
                return visitor.visitChildren(self)




    def addAssetBlock(self):

        localctx = AssetPackParser.AddAssetBlockContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_addAssetBlock)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 6
            self.match(AssetPackParser.ADD_ASSET)
            self.state = 7
            localctx.local = self.file_name()
            self.state = 8
            localctx.swf = self.file_name()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class RootContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def addAssetBlock(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(AssetPackParser.AddAssetBlockContext)
            else:
                return self.getTypedRuleContext(AssetPackParser.AddAssetBlockContext,i)


        def getRuleIndex(self):
            return AssetPackParser.RULE_root

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRoot" ):
                listener.enterRoot(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRoot" ):
                listener.exitRoot(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRoot" ):
                return visitor.visitRoot(self)
            else:
                return visitor.visitChildren(self)




    def root(self):

        localctx = AssetPackParser.RootContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_root)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 13
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==1:
                self.state = 10
                self.addAssetBlock()
                self.state = 15
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class File_nameContext(ParserRuleContext):
        __slots__ = 'parser'

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def FILE_NAME(self):
            return self.getToken(AssetPackParser.FILE_NAME, 0)

        def getRuleIndex(self):
            return AssetPackParser.RULE_file_name

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFile_name" ):
                listener.enterFile_name(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFile_name" ):
                listener.exitFile_name(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFile_name" ):
                return visitor.visitFile_name(self)
            else:
                return visitor.visitChildren(self)




    def file_name(self):

        localctx = AssetPackParser.File_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_file_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 16
            self.match(AssetPackParser.FILE_NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import os
import shutil
from pathlib import Path

def clean_scripts(decomp_location: Path, modified_scripts: set[Path]) -> None:
    """Delete all non-modified scripts.

    Taken from https://stackoverflow.com/questions/19309667/recursive-os-listdir
    - Make recursive os.listdir.
    """
    scripts = [
        Path(dp, f) for dp, _, fn in os.walk(decomp_location.expanduser()) for f in fn
    ]

    for script in scripts:
        if script not in modified_scripts:
            script.unlink()

def copy_file(source: Path, dest: Path) -> None:
    """Copy a file or folder."""
    isdir = source.is_dir()
    if dest.exists():
        if isdir:
            shutil.rmtree(dest)
        else:
            Path.unlink(dest)

    if isdir:
        shutil.copytree(source, dest)
    else:
        shutil.copy(source, dest)



from __future__ import annotations

import os
import subprocess
from logging import info
from pathlib import Path

LOCATION_APT = Path("/usr/bin/ffdec")
LOCATION_FLATPAK = Path("/usr/bin/flatpak")
LOCATION_WINDOWS = Path(f"{os.getenv('PROGRAMFILES')}\\FFDec\\ffdec.bat")
LOCATION_WOW64 = Path(f"{os.getenv('PROGRAMFILES(X86)')}\\FFDec\\ffdec.bat")

ARGS_FLATPAK = [
    "run",
    "--branch=stable",
    "--arch=x86_64",
    "--command=ffdec.sh",
    "com.jpexs.decompiler.flash",
]

class FFDecInterface:
    """An interface to interact with FFDec via the shell.

    This class should only be used for functions that have a one-to-one correspondence
    with FFDec CLI commands.
    For anything more sophisticated, you should use the CompilationManager class instead.
    """

    path: Path
    args: list[str]

    def __init__(
        self: FFDecInterface,
        path: Path | None = None,
        args: list[str] | None = None,
    ) -> None:
        """Initialize by detecting FFDec, or using a provided version"""
        if path is not None:
            self.path = path
            if args is None:
                self.args = []
            else:
                self.args = args

            info("Using FFDec at: %s", path)

        else:
            # Auto-detect FFDec at any of the default locations
            ffdec_installed = any([
                self.install_ffdec(LOCATION_APT),
                self.install_ffdec(LOCATION_FLATPAK, ARGS_FLATPAK),
                self.install_ffdec(LOCATION_WINDOWS),
                self.install_ffdec(LOCATION_WOW64),
            ])

            if not ffdec_installed:
                raise ModuleNotFoundError(
                    """Failed to locate dependency: JPEXS Flash Decompiler.
                    You can download FFDec from this link:
                    https://github.com/jindrapetrik/jpexs-decompiler/releases
                    """
                )

            info("Using FFDec at: %s", self.path)

    def install_ffdec(self: FFDecInterface, path: Path, args: list[str] | None = None) -> bool:
        """Install FFDec from a path. Return true if the installation was successful."""
        if path.exists() and args is None:
            # Normal FFDec install, we're just running ffdec.sh directly
            self.path = path
            self.args = []
            return True

        if path.exists():
            # We're running FFDec through a sandbox or proxy (like Flatpak)
            # path.exists() checks that the path exists, but not that FFDec is installed
            # so we need to run ffdec -help to verify it's installed correctly
            testrun = subprocess.run(
                [path, *args, "-help"],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                check=False,
            )

            if testrun.returncode == 0:
                self.path = path
                self.args = args
                return True

        return False

    def dump_xml(
        self: FFDecInterface,
        inputfile: Path,
        output_dir: Path,
    ) -> bool:
        """Dump XML data of the input file into the output location.

        Returns True if dump was successful.
        """
        process = subprocess.run(
            [self.path, *self.args, "-swf2xml", inputfile, output_dir],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
        )
        return process.returncode == 0

    def rebuild_xml(
        self: FFDecInterface,
        input_dir: Path,
        output_file: Path,
    ) -> bool:
        """Rebuild XML data from a directory into an output SWF file.

        Return True on success.
        """
        process = subprocess.run(
            [self.path, *self.args, "-xml2swf", input_dir, output_file],
            stdout=subprocess.DEVNULL,
            check=False,
        )
        return process.returncode == 0

    def export_scripts(
        self: FFDecInterface,
        inputfile: Path,
        output_dir: Path,
    ) -> bool:
        """Export scripts from a SWF file into a directory.
        
        Returns True on success.
        """
        info("Exporting scripts into %s...", output_dir)

        # set check=False, we will verify the return code manually later
        process = subprocess.run(
            [
                self.path,
                *self.args,
                "-export",
                "script",
                output_dir,
                inputfile,
            ],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=False,
        )

        return process.returncode == 0

    def recompile_data(
        self: FFDecInterface,
        part: str,
        decomp_location: Path,
        swf: Path,
        output: Path,
    ) -> bool:
        """Recompile data of a given type into the SWF file."""

        # Part types: SymbolClass, Movies, Sounds, Shapes, Images, Text, Script
        info("Reimporting %s...", part)

        # set check=False, we will verify the return code manually later
        process = subprocess.run(
            [
                self.path,
                *self.args,
                f"-import{part}",
                swf,
                output,
                decomp_location,
            ],
            stdout=subprocess.DEVNULL,
            check=False,
        )

        return process.returncode == 0



from pathlib import Path

def get_decomp_locations(xml_mode: bool) -> (Path, Path):
    """Return (DECOMP_LOCATION, DECOMP_LOCATION_WITH_SCRIPTS)"""
    if xml_mode:
        return Path("./.Patcher-Temp/swf.xml"), Path("./.Patcher-Temp/")

    return Path("./.Patcher-Temp/mod/"), Path("./.Patcher-Temp/mod/scripts/")
